import numpy as np
import euchar.utils as u
import euchar.filtrations as f
import euchar.cppbinding.curve as cppcurve

#=================================================

def image_2D(image, vector_of_euler_changes_2D=None,
             max_intensity=255):
    """
    Euler characteristic curve of 2D image.
    
    This uses the vector of all possible Euler characteristic changes 
    produced by a pixel insertion in 2D images. This is recomputed 
    every time this function is called if it is not passes as a 
    parameter.

    Parameters
    ----------
    image
        np.ndarray of integers
    vector_of_euler_changes_2D
        list of integers, precomputed Euler characteristic changes
        produced by a single pixel insertion
    max_intensity
        maximum value of elements in image
    
    Return
    ------
    euler_char_curve
        np.array of integers
    """

    try:
        err_line1 = "`image` parameter needs to be an np.ndarray\n---"
        assert str(type(image)) == "<class 'numpy.ndarray'>", err_line1
    except AssertionError as err:
        print("---\nError")
        print(err)

    try:
        err_line1 = "`image` must be a two dimensional np.ndarray\n---"
        assert len(image.shape) == 2, err_line1
    except AssertionError as err:
        print("---\nError")
        print(err)

    if vector_of_euler_changes_2D is None:
        vector_of_euler_changes_2D = u.vector_all_euler_changes_in_2D_images()

    euler_char_curve = cppcurve.image_2d(image,
                                         vector_of_euler_changes_2D,
                                         max_intensity)

    return np.array(euler_char_curve)

#=================================================

def image_3D(image, vector_of_euler_changes_3D=None, max_intensity=255):
    """
    Euler characteristic curve of 3D image.
    
    This uses the vector of all possible Euler characteristic changes 
    produced by a voxel insertion in 3D images. This is must be passed
    as a parameter.
    
    Parameters
    ----------
    image
        np.ndarray of integers
    vector_of_euler_changes_3D
        list of integers, precomputed Euler characteristic changes
        produced by a single voxel insertion
    max_intensity
        maximum value of elements in image
    
    Return
    ------
    euler_char_curve
        np.ndarray of integers

    """

    try:
        err_line1 = "`image` parameter needs to be an np.ndarray\n---"
        assert str(type(image)) == "<class 'numpy.ndarray'>", err_line1
    except AssertionError as err:
        print("---\nError")
        print(err)
        return None

    try:
        err_line1 = "`image` must be a three dimensional np.ndarray\n---"
        assert len(image.shape) == 3, err_line1
    except AssertionError as err:
        print("---\nError")
        print(err)
        return None

    try:
        err_line1 = "You must pass in the vector of all possible Euler\n"
        err_line2 = "characteristic changes for 3D images, which can be\n"
        err_line3 = "computed with euchar.utils.vector_all_euler_changes_in_3D_images.\n---"
        assert vector_of_euler_changes_3D is not None, err_line1 + err_line2 + err_line3
    except AssertionError as err:
        print("---\nError")
        print(err)
        return None

    euler_char_curve = cppcurve.image_3d(image,
                                         vector_of_euler_changes_3D,
                                         max_intensity)

    return np.array(euler_char_curve)

#=================================================

def filtration(simplices, parametrization, bins):
    """
    Compute the Euler characteristic curve of a filtration.
    
    Parameters
    ----------
    simplices
        np.ndarray of shape (N, 3) or (N, 4). Each row is a simplex.
        For example [2 4 -1] is the edge (2,4) and [2 4 9] the 
        triangle (2,4,9). 
    parametrization
        np.ndarray of floats
    bins
        np.ndarray of sorted floats, used to bin the parametrization
        values of simplices 
    
    Return
    ------
    euler_char_curve
        np.ndarray of integers

    """

    try:
        err_line1 = "`simplices` and `parametrization` must have same\n"
        err_line2 = "length."
        assert len(simplices) == len(parametrization), err_line1+err_line2
    except AssertionError as err:
        print("---\nError")
        print(err)
        return None

    dim_simplices = u.simplices_to_dimensions(simplices) 
    euler_char_curve = cppcurve.filtration(dim_simplices, parametrization, bins)
        
    return  np.array(euler_char_curve)



import numpy as np
import matplotlib.pyplot as plt
import euchar.utils as u
from scipy.spatial.distance import euclidean

#=================================================
def matplotlib_plot(rows=1, columns=1, figsize=(4,4), facecolor="w",
                    font_size=10, font_family="CMU Serif"):
    """
    Returns a fig and ax object.
    
    Parameters
    ----------
    rows
        int
    cols
        int
    figsize
        two tuple
    font_size
        int
    facecolor
        string, default 'white'.
    font_family
        string
    
    Returns
    -------
    fig, ax
        matplotlib plots objects. ax will be an array of axes
        if rows>1 and/or cols>1
    
    Example
    -------
    >>> import matplotlib.pyplot as plt
    >>> from euchar.display import matplotlib_plot
    >>> from numpy.random import rand
    >>> points1, points2 = rand(10, 2), rand(1000, 2)
    >>> fig, ax = matplotlib_plot(1, 2)
    >>> ax[0].scatter(points1[:,0], points1[:,1], c='red')
    >>> ax[1].scatter(points2[:,0], points2[:,1], c='blue')
    >>> plt.show()
    
    """

    figsize = (figsize[0]*columns, figsize[1]*rows)
    fig, ax = plt.subplots(rows, columns, figsize=figsize)
    
    # Global Setting
    plt.rcParams['font.family'] = 'CMU Serif'
    plt.rcParams['font.size'] = str(font_size)              
    plt.rcParams['mathtext.fontset'] = 'cm'       
    plt.rcParams['mathtext.rm'] = 'CMU Serif'      
    plt.rcParams['grid.color'] = '#F0F0F0'
    plt.rcParams['grid.linestyle'] = 'solid'
    plt.rcParams['text.usetex'] = True   
    
    fig.patch.set_facecolor(facecolor)
    
    return fig, ax

#=================================================

def visualize_triangle_alpha_parametrization(triangle, points, figsize=(3,3)):
    """
    Visualize the miniball of a triangle in the plane.

    Parameters
    ----------

    triangle
       np.ndarray of three integers, the indices of the vertices of the 
        triangle in `points`
    points
        np.ndarray of shape (N, 2), points in 2 dimensions
    figsize
        two-tuple
    
    """
    
    triangle = np.array(triangle)
    if np.sum(triangle != -1) != 3:
        print(f"`triangle` must be a triangle. Your input is {triangle},")
        print(f"which has {np.sum(triangle != -1)} vertices.")
        return None
    
    def plot_circle(ax, C, r, n=100):
        circle = []
        for i in range(n+1):
            x = r * np.cos(2*np.pi / n * i) + C[0]
            y = r * np.sin(2*np.pi / n * i) + C[1]
            circle.append([x,y])
        circle = np.array(circle)
        ax.plot(circle[:,0], circle[:,1], "-r", alpha=0.6, zorder=1)
    
    tri = points[triangle]
    fig, ax = matplotlib_plot(1, 1, figsize=figsize)
    C = u.center_triangle(*tri)
    r = u.parameter_triangle(*tri) / 2
    plot_circle(ax, C, r, n=100)
    ax.scatter(tri[:,0], tri[:,1])
    ax.scatter(C[0], C[1], c="k")
    fig.tight_layout()
    
#=================================================
    
def visualize_tetrahedron_alpha_parametrization(tetrahedron, points, figsize=(3,3)):
    """Visualize the miniball of a tetrahedron in 3D."""
    
    def plot_sphere(C, r, tetra=None, n=100, figsize=(5,5)):
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111, projection='3d')
    
        print("radius is:", r)
        radius_ok = []
        inside = []
        print("Distances from points to center miniball are:")
        for point in tetra:
            radius_ok.append(np.isclose(r, euclidean(C, point)))
            inside.append(r+0.00000001 >= euclidean(C, point))
            print(euclidean(C, point), end=" | ")
    
        # Make data
        w = np.linspace(0, 2 * np.pi, 100)
        v = np.linspace(0, np.pi, 100)
        x = r * np.outer(np.cos(w), np.sin(v)) + C[0]
        y = r * np.outer(np.sin(w), np.sin(v)) + C[1]
        z = r * np.outer(np.ones(np.size(w)), np.cos(v)) + C[2]

        # Plot the surface
        ax.plot_surface(x, y, z, color='cyan', alpha=0.4)
    
        # Plot tetra vertices
        if tetra is not None:
            ax.scatter(tetra[:,0], tetra[:,1], tetra[:,2], s = 10, c="r")
        #Plot center
        ax.scatter([C[0]], [C[1]], [C[2]], s = 30, c="k")
    
        ax.set(xlim=[-.3, 1.3], ylim=[-.3, 1.3], zlim=[-.3, 1.3])
    
        fig.tight_layout()
    
    tetra = points[tetrahedron]
    r = u.parameter_tetrahedron(*tetra) / 2
    C = u.center_tetrahedron(*tetra)
    plot_sphere(C, r, tetra)

#=================================================

def piecewise_constant_curve(domain, curve):
    """
    Obtain piecewise constant version of a discrete Euler characteristic
    curve.
    """
    new_domain = []
    new_curve = []
    for i, (a, b) in enumerate(zip(domain[:-1], domain[1:])):
        new_domain.extend([a, (b-a)*0.999999 + a])
        new_curve.extend([curve[i], curve[i]])

    new_domain.append(domain[-1])
    new_curve.append(curve[-1])
    
    return np.array(new_domain), np.array(new_curve)

#=================================================

def piecewise_constant_surface(bins1, bins2, surface):
    """
    Obtain the piecewise constant version of a discrete Euler 
    characteristic surface.
    """
    new_bins1, new_bins2 = [], []
    new_surface = np.zeros(shape=(2*len(bins1)-1, 2*len(bins2)-1))

    for i, (a, b) in enumerate(zip(bins1[:-1], bins1[1:])):
        new_bins1.extend([a, (b-a)*0.999999 + a])
    new_bins1.append(bins1[-1])
    
    for j, (c, d) in enumerate(zip(bins2[:-1], bins2[1:])):
        new_bins2.extend([c, (d-c)*0.999999 + c])
    new_bins2.append(bins2[-1])
        
    for i, (a, b) in enumerate(zip(bins1[:-1], bins1[1:])):
        for j, (c, d) in enumerate(zip(bins2[:-1], bins2[1:])):
            val = surface[i][j]
            new_surface[2*i][2*j]     = val
            new_surface[2*i+1][2*j]   = val
            new_surface[2*i][2*j+1]   = val
            new_surface[2*i+1][2*j+1] = val

    final_row = []
    for item in surface[-1,:-1]:
        final_row.extend([item, item])
    final_row.append(surface[-1,-1])

    final_col = []
    for item in surface[:-1,-1]:
        final_col.extend([item, item])
    final_col.append(surface[-1,-1])

    new_surface[-1] = np.array(final_row)
    new_surface[:,-1] = np.array(final_col)

    return np.array(new_bins1), np.array(new_bins2), new_surface
    
#=================================================

def euler_curve_plot(fig, ax, bins, euler_char_curve, 
                     line_color="k", line_width=2,
                     xlabel="Parameter", ylabel="$\chi(K)$", 
                     title="", 
                     xlim=None, ylim=None,
                     size_arrows=None,
                     xticks=None, yticks=None, 
                     xticks_length=None, xticks_width=None,
                     yticks_length=None, yticks_width=None,
                     xticks_locations=None,
                     yticks_locations=None,
                     font_size_ticks=12):
    """
    Display an Euler characteristic curve as a piecewise constant curve. 
    plot.
    
    Parameters
    ----------
    bins,
        np.ndarray, x-axis values of the plot
    euler_char_curve
        np.ndarray of integers
    line_color
       string, default is "k"
    line_width
       int
    xticks, yticks
        lists, ticks to plot along axes
    xlabel, ylabel
        strings, labels of axes
    title
        string
    xlim, ylim
        two-tuples, limits of the plot
    xticks_locations, yticks_locations
        list of two-tuples
    xticks_spacing, yticks_spacing
        two-tuples, delta of space between tick mark and axis
    x_arrow_head_width, x_arrow_head_length
        floats
    y_arrow_head_width, y_arrow_head_length
        floats

    """
    new_bins, new_curve = piecewise_constant_curve(bins, euler_char_curve)
    shape_pixels = fig.get_size_inches()*fig.dpi
    
    ax.plot(new_bins, new_curve, color=line_color,
            linewidth=line_width, marker="", alpha=1, zorder=3)
    
    # limits and title 
    if xlim is None:
        xlim = [min(bins), max(bins)*1.2]
    if ylim is None:
        ylim = [min(euler_char_curve), max(euler_char_curve)*1.2]
    ax.set(xlim=xlim, ylim=ylim, title=title)
    
    # define proportion arrows
    if size_arrows is None:
        size_arrows = int((shape_pixels[0] + shape_pixels[1])/50)
    
    x_arrow_head_width = (ylim[1]-ylim[0]) / shape_pixels[1] * size_arrows / 1.27
    x_arrow_head_length = (xlim[1] - xlim[0]) / shape_pixels[0] * size_arrows * 1.27
    y_arrow_head_width = (xlim[1]-xlim[0]) / shape_pixels[0] * size_arrows / 1.27
    y_arrow_head_length = (ylim[1] - ylim[0]) / shape_pixels[1] * size_arrows* 1.27
    
    # axes off
    ax.axis("off")
    
    
    # x-axis
    ax.arrow(x=xlim[0], y=0, 
             dx=xlim[1] - xlim[0] - x_arrow_head_length, dy=0,
             head_width=x_arrow_head_width,
             head_length=x_arrow_head_length,
             fc="k", zorder=1)
    
    if xticks_length is None:
        xticks_length = (ylim[1] - ylim[0]) / 30
    if xticks_width is None:
        xticks_width = 1
    if xticks_locations is None:
        xticks_locations = [(tick - y_arrow_head_width * 0.8, 
                             -(1.8 * y_arrow_head_length)) for tick in xticks]
        
    for tick, location in zip(xticks, xticks_locations):
        ax.plot([tick, tick], [0, -xticks_length], color="k", linewidth=xticks_width)
        ax.annotate(str(tick), xy=(0,0), xytext=location,
                    fontsize=font_size_ticks)
        
    # y-axis
    ax.arrow(x=bins[0], y=ylim[0], dx=0, 
             dy=ylim[1] - ylim[0] - y_arrow_head_length,
             head_width=y_arrow_head_width,
             head_length=y_arrow_head_length,
             fc="k", zorder=1)
    if yticks_length is None:
        yticks_length = (xlim[1] - xlim[0]) / 30
    if yticks_width is None:
        yticks_width = 1
    if yticks_locations is None:
        yticks_locations = [(-(1.5 * x_arrow_head_length) * len(str(tick)), 
                             tick - x_arrow_head_width * 0.8) for tick in yticks]
    
    for tick, location in zip(yticks, yticks_locations):
        ax.plot([-yticks_length, 0], [tick, tick],
                 color="k", linewidth=xticks_width)
        ax.annotate(str(tick), xy=(0,0), xytext=location,
                   fontsize=font_size_ticks)

#=================================================

def euler_surface_plot(fig, ax, bins1, bins2, euler_char_surf, 
                       n_levels=30, levels=None,
                       min_level=None, max_level=None,
                       xticks=None, yticks=None, colorbar_ticks=None,
                       xlim=None, ylim=None,
                       xlabel="Parametrization 1",
                       ylabel="Parametrization 2", 
                       dx=0.05, dy=0.05,
                       title="", color_map='coolwarm'):
    """
    Display an Euler characteristic surface as a piecewise constant contour 
    plot.
    
    Parameters
    ----------
    bins1, bin2
        np.ndarray, x-axis and y-axis values of the plot
    euler_char_surface
        np.ndarray of integers
    n_levels
        int, number of levels of the contour plot
    min_levels, max_levels
        floats, minimum and maximum level of the contour plot
    xticks, yticks
        lists, ticks to plot along axes
    colorbar_ticks
        list, ticsk of the colorbar
    xlim, ylim
        two-tuples, limits of the plot
    xlabel, ylabel
        strings, labels of axes
    dx, dy
        floats, resolution parameters of the contour plot. 
        Make these smaller to increase resolution.
    title
        string
    color_map
        string, default is 'coolwarm'. Other options are: 'Greys', 
        'PiYG', 'viridis', 'magma', etc.
  
    """
    
    bins1, bins2, euler_char_surf = piecewise_constant_surface(bins1, bins2, euler_char_surf)
    
    # Contours levels
    if min_level is None:
        min_level = euler_char_surf.min()
    if max_level is None:
        max_level = euler_char_surf.max()
    if colorbar_ticks is None:
        colorbar_ticks = [min_level, max_level]
    if levels is None:
        levels = [int(el) for el in np.linspace(min_level, max_level, num=n_levels)]
    
    # Color map
    cmap = plt.get_cmap(color_map)
    
    # Plot the surface.
    xx, yy = np.meshgrid(bins1, bins2)
    cf = ax.contourf(xx + dx/2., yy + dy/2.,
                     euler_char_surf,
                     levels=levels, cmap=cmap)
    fig.colorbar(cf, ax=ax, ticks=colorbar_ticks)
    
    if xticks is None:
        xticks = [min(bins1), max(bins1)]
    if yticks is None:
        yticks = [min(bins2), max(bins2)]
    if xlim is None:
        xim = [min(bins1), max(bins1)]
    if ylim is None:
        ylim = [min(bins2), max(bins2)]
    
    ax.set(title=title,
           xticks=xticks, yticks=yticks,
           xlim=xlim, ylim=ylim,
           xlabel=xlabel, ylabel=ylabel)



import numpy as np
from scipy.spatial.distance import euclidean
from sklearn.neighbors import NearestNeighbors
from euchar.cppbinding.utils import vector_of_euler_changes_2d, vector_of_euler_changes_3d

#=================================================

def estimate_inverse_density(points, n):
    """
    Estimates the inverse of the density at each `p` in `points` as the
    root mean square of the distances from `p` to its `n` nearest 
    neighbors.

    Parameters
    ----------
    points
        np.ndarray of shape (N, d)
    n
        int, number of nearest neighbors used to estimate the inverse of
        the density.
    
    Returns
    -------
    param_vertices
        1-D np.ndarray parametrizing vertices by inverse density.

    """
    nbrs = NearestNeighbors(n+1, algorithm="ball_tree")
    dist = nbrs.fit(points).kneighbors(points)[0][:,1:]

    densities = np.empty(len(points))
    for i in range(len(points)):
        densities[i] = np.sqrt(np.sum((dist[i]**2) / len(dist[i])))
    return densities

#=================================================

def vector_all_euler_changes_in_2D_images():
    """
    Returns a vector of 256 integers, containing all the possible Euler
    characteristic changes produced by the insertion of a central pixel
    in a 3x3 binary neighbourhood.
    """
    
    list_2d_changes = vector_of_euler_changes_2d()
    return np.array(list_2d_changes)

#=================================================

def vector_all_euler_changes_in_3D_images(start=0, end=67_108_864):
    """
    Returns a vector of 67,108,864 integers, containing all the
    possible Euler characteristic changes produced by the insertion of
    a central voxel in a 3x3x3 binary neighbourhood.
    
    Notes
    -----
    It is recommended to produce this vector once and to write it 
    to a file. For example with numpy

    >>> vector_changes = vector_all_euler_changes_in_3d_images()
    >>> np.save("/path/to/dir/changes_3d.npy", vector_changes)
    
    The next time the vector can be loaded with

    >>> vector_changes = np.load("/path/to/dir/changes_3d.npy")
    """
    
    print("\n-----\nIt is recommended to save the result to a file, and to load it when needed.\n-----\n")

    list_3d_changes = vector_of_euler_changes_3d(start, end)
    return np.array(list_3d_changes)

#=================================================

def magnitude(v):
    """Euclidean norm of the vector v."""
    return np.sqrt(v.dot(v))

#=================================================

def circumradius_triangle(A, B, C):
    """
    Compute circumradius of 3 points in 2 or 3 dimensions.
    
    Parameters
    ----------
    A, B, C
        np.arrays of length equal to the Euclidean dimension.

    Return
    ------
    circumradius
        float, circumradius of the 3 points.
    """
    a = np.array(A) - np.array(C)
    b = np.array(B) - np.array(C)
    cross = np.cross(a,b)
    return magnitude(a) * magnitude(b) * magnitude(a-b) \
           / 2 / magnitude(cross)

#=================================================

def circumcenter_triangle(A, B, C):
    """
    Compute circumcenter of 3 points in 2 or 3 dimensions.

    Parameters
    ----------
    A, B, C
        np.arrays of length equal to the Euclidean dimension.

    Return
    ------
    circumcenter
        np.array, the circumcenter of the three points
    """
    a = np.array(A) - np.array(C)
    b = np.array(B) - np.array(C)
    aXb = np.cross(a,b)
    numer1 = magnitude(a)**2 * b - magnitude(b)**2 * a
    numer = np.dot(numer1, b)*a - np.dot(numer1, a)*b
    denom = 2 * (magnitude(aXb)**2)
    return numer / denom + C

#=================================================

def vol_tetra(A, B, C, D):
    """
    Volume of the tetrahedron whose four vertices are the given four 
    points A, B, C, D.

    Parameters
    ----------
    A, B, C, D
        np.arrays, the four vertices of the tetrahedron

    Return
    ------
    volume
        float, the volume of the tetrahedron
    """
    
    a = A - D
    b = B - D
    c = C - D
    volume = abs(np.dot(a, np.cross(b,c))) / 6
    return volume

#=================================================

def circumradius_tetra(A, B, C, D):
    """
    Compute circumradius of 4 points in 3 dimensions.

    Parameters
    ----------
    A, B, C, D
        np.arrays, the four vertices of the tetrahedron

    Return
    ------
    circumradius
        np.array, the circumradius of the four points
    """
    
    e12 = magnitude(A - B) # length edge AB
    o12 = magnitude(C - D) # length edge opposite to AB
    prod12 = e12 * o12
    
    e13 = magnitude(A - C) # same
    o13 = magnitude(B - D)
    prod13 = e13 * o13
    
    e14 = magnitude(A - D) # same
    o14 = magnitude(B - C)
    prod14 = e14 * o14
    
    numer = np.sqrt((prod12 + prod13 + prod14)*
                    (prod12 + prod13 - prod14)*
                    (prod12 - prod13 + prod14)*
                    (-prod12 + prod13 + prod14))
    denom = 24 * vol_tetra(A, B, C, D)
    circumradius = numer / denom
    return circumradius

#=================================================

def circumcenter_tetra(A, B, C, D):
    """
    Compute circumcenter of 4 points in 3 dimensions.

    Parameters
    ----------
    A, B, C, D
        np.arrays, the four vertices of the tetrahedron

    Return
    ------
    circumcenter
        np.array, the circumcenter of the four points
    """
    
    
    colA1, colA2, colA3 = A - D, B - D, C - D
    matrix = np.vstack([colA1, colA2, colA3])
    inverse_matrix = np.linalg.inv(matrix)
    
    B = np.array([A.dot(A) - D.dot(D), B.dot(B) - D.dot(D),
                  C.dot(C) - D.dot(D)]) / 2
    
    circumcenter = [np.dot(row, B) for row in inverse_matrix]
    return circumcenter

#=================================================

def is_acute(lengths):
    """Check if a triangle is acute with Pythgoras theorem."""
    l = sorted(lengths)
    return l[0]*l[0] + l[1]*l[1] >= l[2]*l[2]

#=================================================

def parameter_triangle(A, B, C):
    """
    Compute the Alpha filtration parameter of a triangle of vertices 
    A, B, C.
    
    Parameters
    ----------
    A, B, C
        np.arrays, the three vertices of the triangle

    Return
    ------
    parameter
        float, Alpha filtration parameter of the triangle. This
        is twice the circumradius of the triangle if the triangle is
        acute, or it is the greatest value among the length of edges 
        of the triangle.

    """
    lengths = [euclidean(A, B), euclidean(B, C), euclidean(A, C)]
    acute = is_acute(lengths)
    
    if acute:
        parameter = 2 * circumradius_triangle(A, B, C)
    else:
        parameter = np.array([euclidean(A, B),
                              euclidean(B, C),
                              euclidean(A, C)]).max()
    
    return parameter


#=================================================

def center_triangle(A, B, C):
    lengths = [euclidean(A, B), euclidean(B, C), euclidean(A, C)]
    acute = is_acute(lengths)
    
    if acute:
        C = circumcenter_triangle(A, B, C)
    else:
        ind_par = np.argmax(lengths)
        centers = np.array([(A + B) / 2, 
                            (B + C) / 2,
                            (A + C) / 2])
        C = centers[ind_par]
    
    return C

#=================================================

def is_on_correct_side(A, B, C, test_point, point):
    cr = np.cross(B-A, C-A)
    dot_test = np.dot(cr, test_point-A)
    dot_point = np.dot(cr, point-A)
    
    if dot_test >= 0 and dot_point >= 0:
        return True
    elif dot_test <= 0 and dot_point <= 0:
        return True
    else:
        return False

#=================================================

def circumcenter_is_inside(circum, A, B, C, D):
    if not is_on_correct_side(A, B, C, test_point=D, point=circum):
        return False
    
    if not is_on_correct_side(A, B, D, test_point=C, point=circum):
        return False
    
    if not is_on_correct_side(A, C, D, test_point=B, point=circum):
        return False
    
    if not is_on_correct_side(C, B, D, test_point=A, point=circum):
        return False
    
    return True

#=================================================
    
def parameter_tetrahedron(A, B, C, D):
    """
    Compute the Alpha filtration parameter of a tetrahedron of vertices
    A, B, C, D.
    
    Parameters
    ----------
    A, B, C, D
        np.arrays, the four vertices of the tetrahedron.

    Return
    ------
    parameter
        float, Alpha filtration parameter of the tetrahedron.
    
    """

    circumcenter = circumcenter_tetra(A, B, C, D)

    if circumcenter_is_inside(circumcenter, A, B, C, D):
        return 2 * circumradius_tetra(A, B, C, D)
    else:
        par_ABC = parameter_triangle(A, B, C)
        par_ABD = parameter_triangle(A, B, D)
        par_ACD = parameter_triangle(A, C, D)
        par_BCD = parameter_triangle(B, C, D)
        return max(par_ABC, par_ABD, par_ACD, par_BCD)

#=================================================

def center_tetrahedron(A, B, C, D):
    circum = circumcenter_tetra(A, B, C, D)

    if circumcenter_is_inside(circum, A, B, C, D):
        return circum
    else:
        C1 = center_triangle(A, B, C)
        C2 = center_triangle(A, B, D)
        C3 = center_triangle(A, C, D)
        C4 = center_triangle(B, C, D)
    
        centers = [C1, C2, C3, C4]
    
        par_triABC = parameter_triangle(A, B, C)
        par_triABD = parameter_triangle(A, B, D)
        par_triACD = parameter_triangle(A, C, D)
        par_triBCD = parameter_triangle(B, C, D)
    
        parameters = [par_triABC, par_triABD, par_triACD, par_triBCD]
        ind = np.argmax(np.array(parameters))
        return centers[ind]
    
#=================================================

def simplices_to_dimensions(simplices):
    """
    Transforms an array of simplices to an array of their dimensions.

    Example
    -------
    Given the array of simplices
    
    >>> [[1 -1 -1], [3 4 -1], [10 12 15]]

    the output will be 

    >>> [0 1 2]

    """
    def dim_simplex(sim):
        cnt = 0
        for el in sim:
            if el != -1:
                cnt += 1
            else:
                break
        return cnt - 1

    return np.apply_along_axis(dim_simplex, 1, simplices)



import numpy as np
import euchar.utils as u 
from scipy.spatial import Delaunay
from scipy.spatial.distance import euclidean

#=================================================

def Delaunay_ed_tri_2D(points):
    """Delaunay triangulation simplices of two dimensional points."""
    if points.shape[1] != 2:
        raise ValueError("Points need to be a two dimensional np.ndarray.")

    delaunay_complex = Delaunay(points)
    triangles = np.array([np.sort(simplex) 
                          for simplex in delaunay_complex.simplices])
    masks_edges = np.array([[0,1], [0,2], [1,2]])
    edges = np.unique(np.vstack([t[masks_edges]
                                 for t in triangles]),
                      axis=0)
    
    return edges, triangles

#=================================================

def Delaunay_ed_tri_tetra_3D(points):
    """Delauanay triangulation simplices of three dimensional points."""
    if points.shape[1] !=3:
        raise ValueError("Points need to be a three dimensional np.ndarray.")
    delaunay_complex = Delaunay(points)
    tetrahedra = np.array([np.sort(simplex)
                           for simplex in delaunay_complex.simplices])
    
    masks_edges = np.array([[0,1], [0,2], [0,3],
                            [1,2], [1,3], [2,3]])
    edges = np.unique(np.vstack([tet[masks_edges]
                                 for tet in tetrahedra]),
                      axis=0)
    
    masks_triangles = np.array([[0, 1, 2], [0, 1, 3],
                                [0, 2, 3], [1, 2, 3]])
    triangles = np.unique(np.vstack([tet[masks_triangles]
                                     for tet in tetrahedra]),
                          axis=0)
    
    return edges, triangles, tetrahedra

#=================================================

def alpha_filtration_2D(points):
    """
    Simplices and corresponding parametrization of 2D point cloud.
    
    Parameters
    ----------
    points
        np.ndarray of shape (N, 3)
    
    Returns
    -------
    simplices
        array of integers of shape (N,3) using -1 as placeholder
        Ex. [1,-1,-1] is vertex 1 and [3,5,-1] is edge (3,5).
    parametrization
        array of floats corresponding to the Alpha filtration parameter
        of each simplex.
    
    """

    # Get vertices, and Delaunay simplices
    vertices = np.arange(len(points))
    edges, triangles = Delaunay_ed_tri_2D(points)

    # Make simplices, with -1 as a placeholder
    vertices = np.array([[v, -1, -1] for v in vertices])
    edges = np.array([[ed[0], ed[1], -1] for ed in edges])
    simplices = np.vstack([vertices, edges, triangles])
    
    # Parametrization
    par_vertices = np.zeros(len(points))
    par_edges = np.array([euclidean(points[ed[0]],
                                    points[ed[1]])
                          for ed in edges])
    par_triangles = np.array([u.parameter_triangle(points[tri[0]],
                                                   points[tri[1]],
                                                   points[tri[2]])
                              for tri in triangles])
    parametrization = np.hstack([par_vertices, par_edges,
                                 par_triangles])

    return simplices, parametrization

#=================================================

def alpha_filtration_3D(points):
    """
    Simplices and corresponding parametrization of 3D point cloud.
    
    Parameters
    ----------
    points
        np.ndarray of shape (N, 3)
    
    Returns
    -------
    simplices
        array of integers of shape (N, 4) using -1 as placeholder
        Ex. [1,-1,-1,-1] is vertex 1 and [3,5,-1,-1] is edge (3,5).
    parametrization
        array of floats corresponding to the Alpha filtration parameter
        of each simplex.
    
    """
    # Get vertices, and Delaunay simplices
    vertices = np.arange(len(points))
    edges, triangles, tetrahedra = Delaunay_ed_tri_tetra_3D(points)

    # Make simplices, with -1 as a placeholder
    vertices = np.array([[v, -1, -1, -1] for v in vertices])
    edges = np.array([[ed[0], ed[1], -1, -1] for ed in edges])
    triangles = np.array([[tri[0], tri[1], tri[2], -1]
                          for tri in triangles])
    simplices = np.vstack([vertices, edges, triangles, tetrahedra])
    
    # Parametrization
    par_vertices = np.zeros(len(points))
    par_edges = np.array([euclidean(points[ed[0]],
                                    points[ed[1]])
                          for ed in edges])
    par_triangles = np.array([u.parameter_triangle(points[tri[0]],
                                                   points[tri[1]],
                                                   points[tri[2]])
                              for tri in triangles])
    par_tetrahedra = np.array([u.parameter_tetrahedron(*points[tetra])
                               for tetra in tetrahedra])
    parametrization = np.hstack([par_vertices, par_edges,
                                 par_triangles, par_tetrahedra])

    return simplices, parametrization



    return np.array(parametrization)

#=================================================

def inverse_density_filtration(points, simplices, n_neighbors):
    """
    Parametrization of simplices based on inverse of local density
    at vertices. The inverse of the density at each `p` in `points`
    is estimated as the root mean square of the distances from `p` to
    its `n` nearest neighbors.
    
    Parameters
    ----------
    points
       np.ndarray of shape (N, d)
    simplices
        np.ndarray of shape (M, d+1)
    n_neighbors
        int, number of nearest neighbors of a point `p` in `points`
        used to estimate the local density around `p`.
    
    Return
    ------
    density_parametrization
        1-D np.ndarray, parametrization of `simplices`. The value of a 
        simplex is the maximum of any of its vertices.
    """

    density = u.estimate_inverse_density(points, n_neighbors)

    parametrization = []
    
    for sim in simplices:
        par = max([density[vertex] for vertex in sim if vertex != -1])
        parametrization.append(par)

    return np.array(parametrization)



import os
import re
import sys
import platform
import subprocess

from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext
from distutils.version import LooseVersion


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = sourcedir


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError("CMake must be installed to build the following extensions: " +
                               ", ".join(e.name for e in self.extensions))

        if platform.system() == "Windows":
            cmake_version = LooseVersion(re.search(r'version\s*([\d.]+)', out.decode()).group(1))
            if cmake_version < '3.1.0':
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.sourcedir)))
        cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
                      '-DPYTHON_EXECUTABLE=' + sys.executable]

        cfg = 'Debug' if self.debug else 'Release'
        build_args = ['--config', cfg]

        if platform.system() == "Windows":
            cmake_args += ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(cfg.upper(), extdir)]
            if sys.maxsize > 2**32:
                cmake_args += ['-A', 'x64']
            build_args += ['--', '/m']
        else:
            cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get('CXXFLAGS', ''),
                                                              self.distribution.get_version())
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(['cmake', os.path.abspath(ext.sourcedir)] + cmake_args, cwd=self.build_temp, env=env)
        subprocess.check_call(['cmake', '--build', '.'] + build_args, cwd=self.build_temp)

#===========================================================

with open("README.md", "r") as f:
    long_description = f.read()

utils = CMakeExtension(name="utils",
                       sourcedir="euchar/cppbinding/")

curve = CMakeExtension(name="curve",
                       sourcedir="euchar/cppbinding/")

surface = CMakeExtension(name="surface",
                         sourcedir="euchar/cppbinding/")
setup(
    name='euchar',
    version='0.1.6',
    packages=find_packages("."),
    author="Gabriele Beltramo",
    author_email="gabri.beltramo@gmail.com",
    description="Euler characteristic of images and finite point sets.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gbeltramo/euchar",
    license='MIT',
    install_requires=[
        'numpy',
        'matplotlib',
        'scipy',
        'scikit-learn',
        'pybind11'
    ],
    ext_modules=[utils, curve, surface],
    cmdclass=dict(build_ext=CMakeBuild),
    zip_safe=False,
    python_requires='>=2.7,!=3.1,!=3.2,!=3.3,!=3.4',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Education',
        'Topic :: Scientific/Engineering :: Mathematics',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    keywords='topology data analysis, euler characteristic, euler characteristic curve, euler characteristic surface'
)



import numpy as np
import euchar.utils as u
import euchar.filtrations as f
import euchar.curve as c
from euchar.cppbinding.utils import char_binary_image_2d
import euchar.cppbinding.surface as cppsurface

#=================================================

def images_2D(image1, image2, vector_of_euler_changes_2D=None,
              max_intensity1=255, max_intensity2=255):
    """
    Euler characteristic curve of a pair of 2D images.
    
    This uses the vector of all possible Euler characteristic changes 
    produced by a pixel insertion in 2D images. This is recomputed 
    every time this function is called if it is not passes as a 
    parameter.

    Parameters
    ----------
    image1, image2
        np.ndarray of integers
    vector_of_euler_changes_2D
        list of integers, precomputed Euler characteristic changes
        produced by a single pixel insertion
    max_intensity1, max_intensity2
        maximum value of any input of the form of image1 and image2
    
    Return
    ------
    euler_char_surface
        np.ndarray of integers

    """

    try:
        err_line1 = "`image` parameter needs to be an np.ndarray\n---"
        assert str(type(image1)) == "<class 'numpy.ndarray'>", err_line1
        assert str(type(image2)) == "<class 'numpy.ndarray'>", err_line1
    except AssertionError as err:
        print("---\nError")
        print(err)
        return None

    try:
        err_line1 = "`image` must be a two dimensional np.array\n---"
        assert len(image1.shape) == 2, err_line1
        assert len(image2.shape) == 2, err_line1
    except AssertionError as err:
        print("---\nError")
        print(err)
        return None

    if vector_of_euler_changes_2D is None:
        vector_of_euler_changes_2D = u.vector_all_euler_changes_in_2D_images()

    euler_char_surface = cppsurface.images_2d(image1, image2,
                                            vector_of_euler_changes_2D,
                                            max_intensity1,
                                            max_intensity2)

    return np.array(euler_char_surface)

#=================================================

def image_2D_and_function(image, func, max_intensity=255,
                          iterations=1, kwargs=None):
    """
    Compute the Euler characteristic surface of a give image with the
    bi-filtration given by sublevel sets of pixel values and 
    iterative application of a given function on these sublevel sets.
    
    Parameters
    ----------
    image
        np.ndarray of integers
    iterations
        number of times the function is applied to the image.
    func
        function taking a sublevel set of `image` and returning 
        a binary image of the same size. 
    max_intensity
        maximum value of elements in image
    kwargs
        possible arguments of function `func` passed as a parameter.

    Returns
    -------
    euler_char_surface
        np.ndarray of integers
    
    """
    
    surface = np.zeros(shape=(iterations+1, max_intensity+1), dtype=int)

    surface[0] = c.image_2D(image, max_intensity=max_intensity)
    
    if kwargs is None:
        kwargs = dict()
        
    for i in range(1, iterations+1):
        for j in range(max_intensity+1):
            thresh = (image <= j).astype(np.uint8)
            for _ in range(i):
                thresh = func(thresh, **kwargs)
            euler_char = char_binary_image_2d(thresh.astype(bool))
            surface[i][j] = euler_char
            
    return surface

#=================================================

def images_3D(image1, image2, vector_of_euler_changes_3D=None,
              max_intensity1=255, max_intensity2=255):
    """Euler characteristic surface of 3D images.
    
    This uses the vector of all possible Euler characteristic changes 
    produced by a voxel insertion in 3D images. This is must be passed
    as a parameter.
    
    Parameters
    ----------
    image1, image2
        np.ndarray of integers
    vector_of_euler_changes_3D
        list of integers, precomputed Euler characteristic changes
        produced by a single voxel insertion
    max_intensity1, max_intensity2
        maximum values of elements in image1 and image2
    
    Return
    ------
    euler_char_surface
        np.ndarray of integers

    """

    try:
        err_line1 = "`image` parameter needs to be an np.array\n---"
        assert str(type(image1)) == "<class 'numpy.ndarray'>", err_line1
        assert str(type(image2)) == "<class 'numpy.ndarray'>", err_line1
    except AssertionError as err:
        print("---\nError")
        print(err)
        return None

    try:
        err_line1 = "`image` must be a three dimensional np.array\n---"
        assert len(image1.shape) == 3, err_line1
        assert len(image2.shape) == 3, err_line1        
    except AssertionError as err:
        print("---\nError")
        print(err)
        return None

    try:
        err_line1 = "You must pass in the vector of all possible Euler"
        err_line2 = "characteristic changes for 3D images, which can be"
        err_line3 = "computed with euchar.utils.vector_all_euler_changes_in_3D_images.\n---"
        assert vector_of_euler_changes_3D is not None, err_line1 + err_line2 + err_line3
    except AssertionError as err:
        print("---\nError")
        print(err)
        return None

    euler_char_surface = cppsurface.images_3d(image1, image2,
                                              vector_of_euler_changes_3D,
                                              max_intensity1,
                                              max_intensity2)

    return np.array(euler_char_surface)

#=================================================

def bifiltration(simplices, parametrization1, parametrization2,
                 bins1, bins2):
    """
    Compute Euler characteristic surface of bifiltration.
    
    Parameters
    -----------
    simplices
        np.ndarray of shape (N, 3) or (N, 4). Each row is a simplex.
        For example [2 4 -1] is the edge (2,4) and [2 4 9] the 
        triangle (2,4,9). 
    parametrization1, parametrization2
        np.ndarray of floats
    bins1, bins2
        np.ndarrays of sorted floats, used to bin the parametrization
        values of simplices 
    
    Return
    ------
    euler_char_surface
        np.ndarray of integers

    """
    try:
        err_line1 = "parameters need to be np.ndarrays\n---"
        assert str(type(simplices)) == "<class 'numpy.ndarray'>", err_line1
        assert str(type(parametrization1)) == "<class 'numpy.ndarray'>", err_line1
        assert str(type(parametrization2)) == "<class 'numpy.ndarray'>", err_line1
        assert str(type(bins1)) == "<class 'numpy.ndarray'>", err_line1
        assert str(type(bins2)) == "<class 'numpy.ndarray'>", err_line1
    except AssertionError as err:
        print("---\nError")
        print(err)
        return None

    try:
        err_line1 = "simplices must be a two dimensional np.ndarray\n---"
        assert len(simplices.shape) == 2,  err_line1
    except AssertionError as err:
        print("---\nError")
        print(err)
        return None

    dimensions_simplices = u.simplices_to_dimensions(simplices)
    euler_char_surf = cppsurface.bifiltration(dimensions_simplices,
                                              parametrization1,
                                              parametrization2,
                                              bins1, bins2)
    return np.array(euler_char_surf)
